#include <stdio.h>
#include <string.h>
#include "common_porting.h"
#include "cmsis_os.h"
#include "stm32f4xx_hal.h"
#include "FreeRTOS.h"
#include "bhy2.h"
#include "bhy2_parse.h"
#include "common_algorithm.h"
#include "BHI260_task.h"

#if defined(USE_BHI260)

/*#define UPLOAD_FIRMWARE_TO_FLASH */
#ifdef UPLOAD_FIRMWARE_TO_FLASH
#include "firmware/bhi260ap/BHI260AP_BMM150-flash.fw.h"
#else
#if defined(BHI260AP_BMM150_9DOF)
#include "firmware/bhi260ap/BHI260AP_BMM150.fw.h"//BHI260AP+BMM150, AUX interface
#elif defined(BHI260AP_6DOF)
#include "firmware/bhi260ap/BHI260AP.fw.h"//BHI260AP without BMM150
#endif
#endif


#define WORK_BUFFER_SIZE   2048


struct bhy2_dev bhy2;
uint8_t work_buffer[WORK_BUFFER_SIZE];
volatile uint16_t bhi260_fifo_ready = BHY2_INTERRUPT_READY;


#if defined(BHI260AP_BMM150_9DOF)
static void parse_euler(const struct bhy2_fifo_parse_data_info *callback_info, void *callback_ref);
#endif

#if defined(BHI260AP_6DOF)
static void parse_quaternion(const struct bhy2_fifo_parse_data_info *callback_info, void *callback_ref);
#endif

static void parse_meta_event(const struct bhy2_fifo_parse_data_info *callback_info, void *callback_ref);
static void parse_acc_gyro_mag_3axis_s16(const struct bhy2_fifo_parse_data_info *callback_info, void *callback_ref);
static void print_api_error(int8_t rslt, struct bhy2_dev *dev);
static void upload_firmware(uint8_t boot_stat, struct bhy2_dev *dev);
static void time_to_s_ns(uint64_t time_ticks, uint32_t *s, uint32_t *ns, uint64_t *tns);
char* get_api_error(int8_t error_code);
char* get_sensor_name(uint8_t sensor_id);
char* get_sensor_error_text(uint8_t sensor_error);
static void print_api_error(int8_t rslt, struct bhy2_dev *dev);


extern volatile uint8_t int1_flag;
extern volatile uint8_t int2_flag;
uint8_t acc_accuracy, gyro_accuracy, mag_accuracy;
uint8_t euler_accuracy, gamerv_accuracy;


static void time_to_s_ns(uint64_t time_ticks, uint32_t *s, uint32_t *ns, uint64_t *tns)
{
    *tns = time_ticks * 15625; /* timestamp is now in nanoseconds */
    *s = (uint32_t)(*tns / UINT64_C(1000000000));
    *ns = (uint32_t)(*tns - ((*s) * UINT64_C(1000000000)));
}

#if defined(BHI260AP_BMM150_9DOF)
static void parse_euler(const struct bhy2_fifo_parse_data_info *callback_info, void *callback_ref)
{
	(void)callback_ref;
	struct bhy2_data_orientation data;
	uint32_t s, ns;
	uint8_t *accuracy = (uint8_t*)callback_ref;
	if (callback_info->data_size != 7) /* Check for a valid payload size. Includes sensor ID */
	{
		return;
	}

	bhy2_parse_orientation(callback_info->data_ptr, &data);

	uint64_t timestamp = *callback_info->time_stamp; /* Store the last timestamp */

	timestamp = timestamp * 15625; /* Timestamp is now in nanoseconds */
	s = (uint32_t)(timestamp / UINT64_C(1000000000));
	ns = (uint32_t)(timestamp - (s * UINT64_C(1000000000)));

	if (accuracy)
	{
		PDEBUG("Euler,h,%f,p,%f,r:%f,accuracy,%u\r\n",
		data.heading * 360.0f / 32768.0f,
		data.pitch * 360.0f / 32768.0f,
		data.roll * 360.0f / 32768.0f,
		*accuracy);
	}
	else
	{
		PDEBUG("Euler,h,%f,p,%f,r,%f\r\n",
		data.heading * 360.0f / 32768.0f,
		data.pitch * 360.0f / 32768.0f,
		data.roll * 360.0f / 32768.0f);
	}	
}
#endif

#if defined(BHI260AP_6DOF)
static void parse_quaternion(const struct bhy2_fifo_parse_data_info *callback_info, void *callback_ref)
{
	struct Quaternion q_data;	
    	struct EulerAngles e_data;
	(void)callback_ref;
	struct bhy2_data_quaternion data;
	uint32_t s, ns;
	uint8_t *accuracy = (uint8_t*)callback_ref;
	
	if (callback_info->data_size != 11) /* Check for a valid payload size. Includes sensor ID */
	{
		return;
	}

	bhy2_parse_quaternion(callback_info->data_ptr, &data);

	uint64_t timestamp = *callback_info->time_stamp; /* Store the last timestamp */

	timestamp = timestamp * 15625; /* Timestamp is now in nanoseconds */
	s = (uint32_t)(timestamp / UINT64_C(1000000000));
	ns = (uint32_t)(timestamp - (s * UINT64_C(1000000000)));

	if(accuracy)
	{
		PDEBUG(",Quat,x,%d,y,%d,z,%d,w,%d,accuracy,%u\r\n",data.x, data.y, data.z, data.w, *accuracy);
	}

	/*PDEBUG("SID: %u; T: %u.%09u; x: %f, y: %f, z: %f, w: %f; acc: %.2f\r\n",
		callback_info->sensor_id,
		s,
		ns,
		data.x / 16384.0f,
		data.y / 16384.0f,
		data.z / 16384.0f,
		data.w / 16384.0f,
		((data.accuracy * 180.0f) / 16384.0f) / 3.141592653589793f);*/

	q_data.x = data.x / 16384.0f;
	q_data.y = data.y / 16384.0f;
	q_data.z = data.z / 16384.0f;
	q_data.w = data.w / 16384.0f;	

	e_data = QuaternionToEulerAngles(q_data);
	PDEBUG("The roll : %f, pitch : %f, yaw : %f\r\n", e_data.roll*180/3.14, e_data.pitch*180/3.14, e_data.yaw*180/3.14);
}
#endif

static void parse_acc_gyro_mag_3axis_s16(const struct bhy2_fifo_parse_data_info *callback_info, void *callback_ref)
{
	struct bhy2_data_xyz data;

	uint8_t *accuracy = (uint8_t*)callback_ref;

	if (accuracy)
	{
		bhy2_parse_xyz(callback_info->data_ptr, &data);
		
		if(callback_info->sensor_id == BHY2_SENSOR_ID_ACC_WU)
		{
			PDEBUG(",ACC,x,%d,y,%d,z,%d,accuracy,%u",
				data.x,
				data.y,
				data.z,
				*accuracy);
		}
		else if(callback_info->sensor_id == BHY2_SENSOR_ID_GYRO_WU)
		{
			PDEBUG(",GYRO,x,%d,y,%d,z,%d,accuracy,%u",
				data.x,
				data.y,
				data.z,
				*accuracy);
		}
		else if(callback_info->sensor_id == BHY2_SENSOR_ID_MAG_WU)
		{
			PDEBUG(",MAG,x,%d,y,%d,z,%d,accuracy,%u,",
				data.x,
				data.y,
				data.z,
				*accuracy);
		}

	}
	else
	{
		PDEBUG("Null reference\r\r\n");
	}
}

/* BHY2_SYS_ID_META_EVENT, BHY2_SYS_ID_META_EVENT_WU */
void parse_meta_event(const struct bhy2_fifo_parse_data_info *callback_info, void *callback_ref)
{
	(void)callback_ref;
	uint8_t meta_event_type = callback_info->data_ptr[0];
	uint8_t byte1 = callback_info->data_ptr[1];
	uint8_t byte2 = callback_info->data_ptr[2];
	uint8_t *accuracy = (uint8_t*)callback_ref;
	char *event_text;

	if (callback_info->sensor_id == BHY2_SYS_ID_META_EVENT)
	{
		event_text = "[META EVENT]";
	}
	else if (callback_info->sensor_id == BHY2_SYS_ID_META_EVENT_WU)
	{
		event_text = "[META EVENT WAKE UP]";
	}
	else
	{
		return;
	}

	switch (meta_event_type)
	{
		case BHY2_META_EVENT_FLUSH_COMPLETE:
			PDEBUG("%s Flush complete for sensor id %u\r\n", event_text, byte1);
			break;
		case BHY2_META_EVENT_SAMPLE_RATE_CHANGED:
			PDEBUG("%s Sample rate changed for sensor id %u\r\n", event_text, byte1);
			break;
		case BHY2_META_EVENT_POWER_MODE_CHANGED:
			PDEBUG("%s Power mode changed for sensor id %u\r\n", event_text, byte1);
			break;
		case BHY2_META_EVENT_ALGORITHM_EVENTS:
			PDEBUG("%s Algorithm event\r\n", event_text);
			break;
		case BHY2_META_EVENT_SENSOR_STATUS:
			PDEBUG("%s Accuracy for sensor id %u changed to %u\r\n", event_text, byte1, byte2);
			if (accuracy)
			{
				if(byte1 == BHY2_SENSOR_ID_ACC_WU)
				{
					acc_accuracy = byte2;
				}
				else if(byte1 == BHY2_SENSOR_ID_GYRO_WU)
				{
					gyro_accuracy = byte2;
				}
				else if(byte1 == BHY2_SENSOR_ID_MAG_WU)
				{
					mag_accuracy = byte2;
				}
				else if(byte1 == BHY2_SENSOR_ID_ORI_WU)
				{
					euler_accuracy = byte2;
				}
				else if(byte1 == BHY2_SENSOR_ID_GAMERV)
				{
					gamerv_accuracy = byte2;
				}
				else
				{
							*accuracy = byte2;
				}
			}
			break;
		case BHY2_META_EVENT_BSX_DO_STEPS_MAIN:
			PDEBUG("%s BSX event (do steps main)\r\n", event_text);
			break;
		case BHY2_META_EVENT_BSX_DO_STEPS_CALIB:
			PDEBUG("%s BSX event (do steps calib)\r\n", event_text);
			break;
		case BHY2_META_EVENT_BSX_GET_OUTPUT_SIGNAL:
			PDEBUG("%s BSX event (get output signal)\r\n", event_text);
			break;
		case BHY2_META_EVENT_SENSOR_ERROR:
			PDEBUG("%s Sensor id %u reported error 0x%02X\r\n", event_text, byte1, byte2);
			break;
		case BHY2_META_EVENT_FIFO_OVERFLOW:
			PDEBUG("%s FIFO overflow\r\n", event_text);
			break;
		case BHY2_META_EVENT_DYNAMIC_RANGE_CHANGED:
			PDEBUG("%s Dynamic range changed for sensor id %u\r\n", event_text, byte1);
			break;
		case BHY2_META_EVENT_FIFO_WATERMARK:
			PDEBUG("%s FIFO watermark reached\r\n", event_text);
			break;
		case BHY2_META_EVENT_INITIALIZED:
			PDEBUG("%s Firmware initialized. Firmware version %u\r\n", event_text, ((uint16_t)byte2 << 8) | byte1);
			break;
		case BHY2_META_TRANSFER_CAUSE:
			PDEBUG("%s Transfer cause for sensor id %u\r\n", event_text, byte1);
			break;
		case BHY2_META_EVENT_SENSOR_FRAMEWORK:
			PDEBUG("%s Sensor framework event for sensor id %u\r\n", event_text, byte1);
			break;
		case BHY2_META_EVENT_RESET:
			PDEBUG("%s Reset event\r\n", event_text);
			break;
		case BHY2_META_EVENT_SPACER:
			break;
		default:
			PDEBUG("%s Unknown meta event with id: %u\r\n", event_text, meta_event_type);
			break;
	}
}

static void upload_firmware(uint8_t boot_stat, struct bhy2_dev *dev)
{
	uint8_t sensor_error;
	int8_t temp_rslt;
	int8_t rslt = BHY2_OK;

#ifdef UPLOAD_FIRMWARE_TO_FLASH
	if (boot_stat & BHY2_BST_FLASH_DETECTED)
	{
		uint32_t start_addr = BHY2_FLASH_SECTOR_START_ADDR;
		uint32_t end_addr = start_addr + sizeof(bhy2_firmware_image);
		PDEBUG("Flash detected. Erasing flash to upload firmware\r\n");

		rslt = bhy2_erase_flash(start_addr, end_addr, dev);
		print_api_error(rslt, dev);
	}
	else
	{
		printf("Flash not detected\r\n");

		rslt = BHY2_E_IO;
		print_api_error(rslt, dev);
	}

	PDEBUG("Loading firmware into FLASH.\r\n");
	rslt = bhy2_upload_firmware_to_flash(bhy2_firmware_image, sizeof(bhy2_firmware_image), dev);
#else
	PDEBUG("Loading firmware into RAM.\r\n");
	rslt = bhy2_upload_firmware_to_ram(bhy2_firmware_image, sizeof(bhy2_firmware_image), dev);
#endif
	temp_rslt = bhy2_get_error_value(&sensor_error, dev);
	if (sensor_error)
	{
		PDEBUG("%s\r\n", get_sensor_error_text(sensor_error));
	}

	print_api_error(rslt, dev);
	print_api_error(temp_rslt, dev);

#ifdef UPLOAD_FIRMWARE_TO_FLASH
	PDEBUG("Booting from FLASH.\r\n");
	rslt = bhy2_boot_from_flash(dev);
#else
	PDEBUG("Booting from RAM.\r\n");
	rslt = bhy2_boot_from_ram(dev);
#endif

	temp_rslt = bhy2_get_error_value(&sensor_error, dev);
	if (sensor_error)
	{
		PDEBUG("%s\r\n", get_sensor_error_text(sensor_error));
	}

	print_api_error(rslt, dev);
	print_api_error(temp_rslt, dev);
}


char* get_sensor_error_text(uint8_t sensor_error)
{
    char *ret;

    switch (sensor_error)
    {
        case 0x00:
            break;
        case 0x10:
            ret = "[Sensor error] Bootloader reports: Firmware Expected Version Mismatch";
            break;
        case 0x11:
            ret = "[Sensor error] Bootloader reports: Firmware Upload Failed: Bad Header CRC";
            break;
        case 0x12:
            ret = "[Sensor error] Bootloader reports: Firmware Upload Failed: SHA Hash Mismatch";
            break;
        case 0x13:
            ret = "[Sensor error] Bootloader reports: Firmware Upload Failed: Bad Image CRC";
            break;
        case 0x14:
            ret = "[Sensor error] Bootloader reports: Firmware Upload Failed: ECDSA Signature Verification Failed";
            break;
        case 0x15:
            ret = "[Sensor error] Bootloader reports: Firmware Upload Failed: Bad Public Key CRC";
            break;
        case 0x16:
            ret = "[Sensor error] Bootloader reports: Firmware Upload Failed: Signed Firmware Required";
            break;
        case 0x17:
            ret = "[Sensor error] Bootloader reports: Firmware Upload Failed: FW Header Missing";
            break;
        case 0x19:
            ret = "[Sensor error] Bootloader reports: Unexpected Watchdog Reset";
            break;
        case 0x1A:
            ret = "[Sensor error] ROM Version Mismatch";
            break;
        case 0x1B:
            ret = "[Sensor error] Bootloader reports: Fatal Firmware Error";
            break;
        case 0x1C:
            ret = "[Sensor error] Chained Firmware Error: Next Payload Not Found";
            break;
        case 0x1D:
            ret = "[Sensor error] Chained Firmware Error: Payload Not Valid";
            break;
        case 0x1E:
            ret = "[Sensor error] Chained Firmware Error: Payload Entries Invalid";
            break;
        case 0x1F:
            ret = "[Sensor error] Bootloader reports: Bootloader Error: OTP CRC Invalid";
            break;
        case 0x20:
            ret = "[Sensor error] Firmware Init Failed";
            break;
        case 0x21:
            ret = "[Sensor error] Sensor Init Failed: Unexpected Device ID";
            break;
        case 0x22:
            ret = "[Sensor error] Sensor Init Failed: No Response from Device";
            break;
        case 0x23:
            ret = "[Sensor error] Sensor Init Failed: Unknown";
            break;
        case 0x24:
            ret = "[Sensor error] Sensor Error: No Valid Data";
            break;
        case 0x25:
            ret = "[Sensor error] Slow Sample Rate";
            break;
        case 0x26:
            ret = "[Sensor error] Data Overflow (saturated sensor data)";
            break;
        case 0x27:
            ret = "[Sensor error] Stack Overflow";
            break;
        case 0x28:
            ret = "[Sensor error] Insufficient Free RAM";
            break;
        case 0x29:
            ret = "[Sensor error] Sensor Init Failed: Driver Parsing Error";
            break;
        case 0x2A:
            ret = "[Sensor error] Too Many RAM Banks Required";
            break;
        case 0x2B:
            ret = "[Sensor error] Invalid Event Specified";
            break;
        case 0x2C:
            ret = "[Sensor error] More than 32 On Change";
            break;
        case 0x2D:
            ret = "[Sensor error] Firmware Too Large";
            break;
        case 0x2F:
            ret = "[Sensor error] Invalid RAM Banks";
            break;
        case 0x30:
            ret = "[Sensor error] Math Error";
            break;
        case 0x40:
            ret = "[Sensor error] Memory Error";
            break;
        case 0x41:
            ret = "[Sensor error] SWI3 Error";
            break;
        case 0x42:
            ret = "[Sensor error] SWI4 Error";
            break;
        case 0x43:
            ret = "[Sensor error] Illegal Instruction Error";
            break;
        case 0x44:
            ret = "[Sensor error] Bootloader reports: Unhandled Interrupt Error / Exception / Postmortem Available";
            break;
        case 0x45:
            ret = "[Sensor error] Invalid Memory Access";
            break;
        case 0x50:
            ret = "[Sensor error] Algorithm Error: BSX Init";
            break;
        case 0x51:
            ret = "[Sensor error] Algorithm Error: BSX Do Step";
            break;
        case 0x52:
            ret = "[Sensor error] Algorithm Error: Update Sub";
            break;
        case 0x53:
            ret = "[Sensor error] Algorithm Error: Get Sub";
            break;
        case 0x54:
            ret = "[Sensor error] Algorithm Error: Get Phys";
            break;
        case 0x55:
            ret = "[Sensor error] Algorithm Error: Unsupported Phys Rate";
            break;
        case 0x56:
            ret = "[Sensor error] Algorithm Error: Cannot find BSX Driver";
            break;
        case 0x60:
            ret = "[Sensor error] Sensor Self-Test Failure";
            break;
        case 0x61:
            ret = "[Sensor error] Sensor Self-Test X Axis Failure";
            break;
        case 0x62:
            ret = "[Sensor error] Sensor Self-Test Y Axis Failure";
            break;
        case 0x64:
            ret = "[Sensor error] Sensor Self-Test Z Axis Failure";
            break;
        case 0x65:
            ret = "[Sensor error] FOC Failure";
            break;
        case 0x66:
            ret = "[Sensor error] Sensor Busy";
            break;
        case 0x6F:
            ret = "[Sensor error] Self-Test or FOC Test Unsupported";
            break;
        case 0x72:
            ret = "[Sensor error] No Host Interrupt Set";
            break;
        case 0x73:
            ret = "[Sensor error] Event ID Passed to Host Interface Has No Known Size";
            break;
        case 0x75:
            ret = "[Sensor error] Host Download Channel Underflow (Host Read Too Fast)";
            break;
        case 0x76:
            ret = "[Sensor error] Host Upload Channel Overflow (Host Wrote Too Fast)";
            break;
        case 0x77:
            ret = "[Sensor error] Host Download Channel Empty";
            break;
        case 0x78:
            ret = "[Sensor error] DMA Error";
            break;
        case 0x79:
            ret = "[Sensor error] Corrupted Input Block Chain";
            break;
        case 0x7A:
            ret = "[Sensor error] Corrupted Output Block Chain";
            break;
        case 0x7B:
            ret = "[Sensor error] Buffer Block Manager Error";
            break;
        case 0x7C:
            ret = "[Sensor error] Input Channel Not Word Aligned";
            break;
        case 0x7D:
            ret = "[Sensor error] Too Many Flush Events";
            break;
        case 0x7E:
            ret = "[Sensor error] Unknown Host Channel Error";
            break;
        case 0x81:
            ret = "[Sensor error] Decimation Too Large";
            break;
        case 0x90:
            ret = "[Sensor error] Master SPI/I2C Queue Overflow";
            break;
        case 0x91:
            ret = "[Sensor error] SPI/I2C Callback Error";
            break;
        case 0xA0:
            ret = "[Sensor error] Timer Scheduling Error";
            break;
        case 0xB0:
            ret = "[Sensor error] Invalid GPIO for Host IRQ";
            break;
        case 0xB1:
            ret = "[Sensor error] Error Sending Initialized Meta Events";
            break;
        case 0xC0:
            ret = "[Sensor error] Bootloader reports: Command Error";
            break;
        case 0xC1:
            ret = "[Sensor error] Bootloader reports: Command Too Long";
            break;
        case 0xC2:
            ret = "[Sensor error] Bootloader reports: Command Buffer Overflow";
            break;
        case 0xD0:
            ret = "[Sensor error] User Mode Error: Sys Call Invalid";
            break;
        case 0xD1:
            ret = "[Sensor error] User Mode Error: Trap Invalid";
            break;
        case 0xE1:
            ret = "[Sensor error] Firmware Upload Failed: Firmware header corrupt";
            break;
        case 0xE2:
            ret = "[Sensor error] Sensor Data Injection: Invalid input stream";
            break;
        default:
            ret = "[Sensor error] Unknown error code";
    }

    return ret;
}

char* get_sensor_name(uint8_t sensor_id)
{
    char *ret;

    switch (sensor_id)
    {
        case BHY2_SENSOR_ID_ACC_PASS:
            ret = "Accelerometer passthrough";
            break;
        case BHY2_SENSOR_ID_ACC_RAW:
            ret = "Accelerometer uncalibrated";
            break;
        case BHY2_SENSOR_ID_ACC:
            ret = "Accelerometer corrected";
            break;
        case BHY2_SENSOR_ID_ACC_BIAS:
            ret = "Accelerometer offset";
            break;
        case BHY2_SENSOR_ID_ACC_WU:
            ret = "Accelerometer corrected wake up";
            break;
        case BHY2_SENSOR_ID_ACC_RAW_WU:
            ret = "Accelerometer uncalibrated wake up";
            break;
        case BHY2_SENSOR_ID_GYRO_PASS:
            ret = "Gyroscope passthrough";
            break;
        case BHY2_SENSOR_ID_GYRO_RAW:
            ret = "Gyroscope uncalibrated";
            break;
        case BHY2_SENSOR_ID_GYRO:
            ret = "Gyroscope corrected";
            break;
        case BHY2_SENSOR_ID_GYRO_BIAS:
            ret = "Gyroscope offset";
            break;
        case BHY2_SENSOR_ID_GYRO_WU:
            ret = "Gyroscope wake up";
            break;
        case BHY2_SENSOR_ID_GYRO_RAW_WU:
            ret = "Gyroscope uncalibrated wake up";
            break;
        case BHY2_SENSOR_ID_MAG_PASS:
            ret = "Magnetometer passthrough";
            break;
        case BHY2_SENSOR_ID_MAG_RAW:
            ret = "Magnetometer uncalibrated";
            break;
        case BHY2_SENSOR_ID_MAG:
            ret = "Magnetometer corrected";
            break;
        case BHY2_SENSOR_ID_MAG_BIAS:
            ret = "Magnetometer offset";
            break;
        case BHY2_SENSOR_ID_MAG_WU:
            ret = "Magnetometer wake up";
            break;
        case BHY2_SENSOR_ID_MAG_RAW_WU:
            ret = "Magnetometer uncalibrated wake up";
            break;
        case BHY2_SENSOR_ID_GRA:
            ret = "Gravity vector";
            break;
        case BHY2_SENSOR_ID_GRA_WU:
            ret = "Gravity vector wake up";
            break;
        case BHY2_SENSOR_ID_LACC:
            ret = "Linear acceleration";
            break;
        case BHY2_SENSOR_ID_LACC_WU:
            ret = "Linear acceleration wake up";
            break;
        case BHY2_SENSOR_ID_RV:
            ret = "Rotation vector";
            break;
        case BHY2_SENSOR_ID_RV_WU:
            ret = "Rotation vector wake up";
            break;
        case BHY2_SENSOR_ID_GAMERV:
            ret = "Game rotation vector";
            break;
        case BHY2_SENSOR_ID_GAMERV_WU:
            ret = "Game rotation vector wake up";
            break;
        case BHY2_SENSOR_ID_GEORV:
            ret = "Geo-magnetic rotation vector";
            break;
        case BHY2_SENSOR_ID_GEORV_WU:
            ret = "Geo-magnetic rotation vector wake up";
            break;
        case BHY2_SENSOR_ID_ORI:
            ret = "Orientation";
            break;
        case BHY2_SENSOR_ID_ORI_WU:
            ret = "Orientation wake up";
            break;
        case BHY2_SENSOR_ID_TILT_DETECTOR:
            ret = "Tilt detector";
            break;
        case BHY2_SENSOR_ID_STD:
            ret = "Step detector";
            break;
        case BHY2_SENSOR_ID_STC:
            ret = "Step counter";
            break;
        case BHY2_SENSOR_ID_STC_WU:
            ret = "Step counter wake up";
            break;
        case BHY2_SENSOR_ID_SIG:
            ret = "Significant motion";
            break;
        case BHY2_SENSOR_ID_WAKE_GESTURE:
            ret = "Wake gesture";
            break;
        case BHY2_SENSOR_ID_GLANCE_GESTURE:
            ret = "Glance gesture";
            break;
        case BHY2_SENSOR_ID_PICKUP_GESTURE:
            ret = "Pickup gesture";
            break;
        case BHY2_SENSOR_ID_AR:
            ret = "Activity recognition";
            break;
        case BHY2_SENSOR_ID_WRIST_TILT_GESTURE:
            ret = "Wrist tilt gesture";
            break;
        case BHY2_SENSOR_ID_DEVICE_ORI:
            ret = "Device orientation";
            break;
        case BHY2_SENSOR_ID_DEVICE_ORI_WU:
            ret = "Device orientation wake up";
            break;
        case BHY2_SENSOR_ID_STATIONARY_DET:
            ret = "Stationary detect";
            break;
        case BHY2_SENSOR_ID_MOTION_DET:
            ret = "Motion detect";
            break;
        case BHY2_SENSOR_ID_ACC_BIAS_WU:
            ret = "Accelerometer offset wake up";
            break;
        case BHY2_SENSOR_ID_GYRO_BIAS_WU:
            ret = "Gyroscope offset wake up";
            break;
        case BHY2_SENSOR_ID_MAG_BIAS_WU:
            ret = "Magnetometer offset wake up";
            break;
        case BHY2_SENSOR_ID_STD_WU:
            ret = "Step detector wake up";
            break;
        case BHY2_SENSOR_ID_TEMP:
            ret = "Temperature";
            break;
        case BHY2_SENSOR_ID_BARO:
            ret = "Barometer";
            break;
        case BHY2_SENSOR_ID_HUM:
            ret = "Humidity";
            break;
        case BHY2_SENSOR_ID_GAS:
            ret = "Gas";
            break;
        case BHY2_SENSOR_ID_TEMP_WU:
            ret = "Temperature wake up";
            break;
        case BHY2_SENSOR_ID_BARO_WU:
            ret = "Barometer wake up";
            break;
        case BHY2_SENSOR_ID_HUM_WU:
            ret = "Humidity wake up";
            break;
        case BHY2_SENSOR_ID_GAS_WU:
            ret = "Gas wake up";
            break;
        case BHY2_SENSOR_ID_STC_HW:
            ret = "Hardware Step counter";
            break;
        case BHY2_SENSOR_ID_STD_HW:
            ret = "Hardware Step detector";
            break;
        case BHY2_SENSOR_ID_SIG_HW:
            ret = "Hardware Significant motion";
            break;
        case BHY2_SENSOR_ID_STC_HW_WU:
            ret = "Hardware Step counter wake up";
            break;
        case BHY2_SENSOR_ID_STD_HW_WU:
            ret = "Hardware Step detector wake up";
            break;
        case BHY2_SENSOR_ID_SIG_HW_WU:
            ret = "Hardware Significant motion wake up";
            break;
        case BHY2_SENSOR_ID_ANY_MOTION:
            ret = "Any motion";
            break;
        case BHY2_SENSOR_ID_ANY_MOTION_WU:
            ret = "Any motion wake up";
            break;
        case BHY2_SENSOR_ID_EXCAMERA:
            ret = "External camera trigger";
            break;
        case BHY2_SENSOR_ID_GPS:
            ret = "GPS";
            break;
        case BHY2_SENSOR_ID_LIGHT:
            ret = "Light";
            break;
        case BHY2_SENSOR_ID_PROX:
            ret = "Proximity";
            break;
        case BHY2_SENSOR_ID_LIGHT_WU:
            ret = "Light wake up";
            break;
        case BHY2_SENSOR_ID_PROX_WU:
            ret = "Proximity wake up";
            break;
        default:
            if ((sensor_id >= BHY2_SENSOR_ID_CUSTOM_START) && (sensor_id <= BHY2_SENSOR_ID_CUSTOM_END))
            {
                ret = "Custom sensor ID ";
            }
            else
            {
                ret = "Undefined sensor ID ";
            }
    }

    return ret;
}

char* get_api_error(int8_t error_code)
{
	char *ret = " ";

	switch (error_code)
	{
		case BHY2_OK:
			break;
		case BHY2_E_NULL_PTR:
			ret = "[API Error] Null pointer";
			break;
		case BHY2_E_INVALID_PARAM:
			ret = "[API Error] Invalid parameter";
			break;
		case BHY2_E_IO:
			ret = "[API Error] IO error";
			break;
		case BHY2_E_MAGIC:
			ret = "[API Error] Invalid firmware";
			break;
		case BHY2_E_TIMEOUT:
			ret = "[API Error] Timed out";
			break;
		case BHY2_E_BUFFER:
			ret = "[API Error] Invalid buffer";
			break;
		case BHY2_E_INVALID_FIFO_TYPE:
			ret = "[API Error] Invalid FIFO type";
			break;
		case BHY2_E_INVALID_EVENT_SIZE:
			ret = "[API Error] Invalid Event size";
			break;
		case BHY2_E_PARAM_NOT_SET:
			ret = "[API Error] Parameter not set";
			break;
		default:
			ret = "[API Error] Unknown API error code";
	}

	return ret;
}


static void print_api_error(int8_t rslt, struct bhy2_dev *dev)
{
	if (rslt != BHY2_OK)
	{
		PDEBUG("%s\r\n", get_api_error(rslt));
	}
}

int get_interrupt_status(void)
{
	return Get_MCU_INT_Pin_Status();
}

int demo_sensor(void)
{
	uint8_t product_id = 0;
	uint16_t version = 0;
	int8_t rslt;
	int64_t system_current_time = 0;
	uint32_t tick;

	uint8_t hintr_ctrl, hif_ctrl, boot_status;
	uint8_t accuracy; /* Accuracy is reported as a meta event. It is being printed alongside the data */

	float sample_rate = 25.0; /* Read out data measured at 10Hz */
	uint32_t report_latency_ms = 0; /* Report immediately */

	#if defined(USE_I2C_INTERFACE)
	rslt = bhy2_init(BHY2_I2C_INTERFACE, bhy2_i2c_read, bhy2_i2c_write, bst_delay_us, BHY2_RD_WR_LEN, NULL, &bhy2);
	#elif defined(USE_SPI_INTERFACE)
	rslt = bhy2_init(BHY2_SPI_INTERFACE, bhy2_spi_read, bhy2_spi_write, bst_delay_us, BHY2_RD_WR_LEN, NULL, &bhy2);
	#endif
	print_api_error(rslt, &bhy2);

	rslt = bhy2_soft_reset(&bhy2);
	print_api_error(rslt, &bhy2);
	HAL_Delay(100);

	rslt = bhy2_get_product_id(&product_id, &bhy2);
	print_api_error(rslt, &bhy2);

	/* Check for a valid product ID */
	if (product_id != BHY2_PRODUCT_ID)
	{
		PDEBUG("Product ID read %X. Expected %X\r\n", product_id, BHY2_PRODUCT_ID);
	}
	else
	{
		PDEBUG("BHI260/BHA260 found. Product ID read %X\r\n", product_id);
	}

	/* Check the interrupt pin and FIFO configurations. Disable status and debug */
	hintr_ctrl = BHY2_ICTL_DISABLE_STATUS_FIFO | BHY2_ICTL_DISABLE_DEBUG;

	rslt = bhy2_set_host_interrupt_ctrl(hintr_ctrl, &bhy2);
	print_api_error(rslt, &bhy2);
	rslt = bhy2_get_host_interrupt_ctrl(&hintr_ctrl, &bhy2);
	print_api_error(rslt, &bhy2);

	PDEBUG("Host interrupt control\r\n");
	PDEBUG("	Wake up FIFO %s.\r\n", (hintr_ctrl & BHY2_ICTL_DISABLE_FIFO_W) ? "disabled" : "enabled");
	PDEBUG("	Non wake up FIFO %s.\r\n", (hintr_ctrl & BHY2_ICTL_DISABLE_FIFO_NW) ? "disabled" : "enabled");
	PDEBUG("	Status FIFO %s.\r\n", (hintr_ctrl & BHY2_ICTL_DISABLE_STATUS_FIFO) ? "disabled" : "enabled");
	PDEBUG("	Debugging %s.\r\n", (hintr_ctrl & BHY2_ICTL_DISABLE_DEBUG) ? "disabled" : "enabled");
	PDEBUG("	Fault %s.\r\n", (hintr_ctrl & BHY2_ICTL_DISABLE_FAULT) ? "disabled" : "enabled");
	PDEBUG("	Interrupt is %s.\r\n", (hintr_ctrl & BHY2_ICTL_ACTIVE_LOW) ? "active low" : "active high");
	PDEBUG("	Interrupt is %s triggered.\r\n", (hintr_ctrl & BHY2_ICTL_EDGE) ? "pulse" : "level");
	PDEBUG("	Interrupt pin drive is %s.\r\n", (hintr_ctrl & BHY2_ICTL_OPEN_DRAIN) ? "open drain" : "push-pull");

	/* Configure the host interface */
	hif_ctrl = 0;
	rslt = bhy2_set_host_intf_ctrl(hif_ctrl, &bhy2);
	print_api_error(rslt, &bhy2);

	/* Check if the sensor is ready to load firmware */
	rslt = bhy2_get_boot_status(&boot_status, &bhy2);
	print_api_error(rslt, &bhy2);

	if (boot_status & BHY2_BST_HOST_INTERFACE_READY)
	{
		upload_firmware(boot_status, &bhy2);

		rslt = bhy2_get_kernel_version(&version, &bhy2);
		print_api_error(rslt, &bhy2);
		if ((rslt == BHY2_OK) && (version != 0))
		{
			PDEBUG("Boot successful. Kernel version %u.\r\n", version);
		}

		//Do axis remapping if need
		#if 0
		struct bhy2_orient_matrix acc_matrix = {1, 0, 0, 0, 1, 0, 0, 0, 1};
		struct bhy2_orient_matrix gyro_matrix = {1, 0, 0, 0, 1, 0, 0, 0, 1};
		struct bhy2_orient_matrix mag_matrix = {1, 0, 0, 0, -1, 0, 0, 0, -1};
		bhy2_set_orientation_matrix(BHY2_SENSOR_ID_ACC, acc_matrix, &bhy2);
		bhy2_set_orientation_matrix(BHY2_SENSOR_ID_GYRO, gyro_matrix, &bhy2);
		bhy2_set_orientation_matrix(BHY2_SENSOR_ID_MAG, mag_matrix, &bhy2);
		#endif

		rslt = bhy2_register_fifo_parse_callback(BHY2_SYS_ID_META_EVENT, parse_meta_event, (void*)&accuracy, &bhy2);
		print_api_error(rslt, &bhy2);
		rslt = bhy2_register_fifo_parse_callback(BHY2_SYS_ID_META_EVENT_WU, parse_meta_event, (void*)&accuracy, &bhy2);
		print_api_error(rslt, &bhy2);

		rslt = bhy2_register_fifo_parse_callback(BHY2_SENSOR_ID_ACC_WU, parse_acc_gyro_mag_3axis_s16, (void*)&acc_accuracy, &bhy2);
		print_api_error(rslt, &bhy2);

		rslt = bhy2_register_fifo_parse_callback(BHY2_SENSOR_ID_GYRO_WU, parse_acc_gyro_mag_3axis_s16, (void*)&gyro_accuracy, &bhy2);
		print_api_error(rslt, &bhy2);
		
		#if defined(BHI260AP_BMM150_9DOF)
		rslt = bhy2_register_fifo_parse_callback(BHY2_SENSOR_ID_ORI_WU, parse_euler, (void*)&euler_accuracy, &bhy2);
		print_api_error(rslt, &bhy2);

		rslt = bhy2_register_fifo_parse_callback(BHY2_SENSOR_ID_MAG_WU, parse_acc_gyro_mag_3axis_s16, (void*)&mag_accuracy, &bhy2);
		print_api_error(rslt, &bhy2);
		#elif defined(BHI260AP_6DOF)
		rslt = bhy2_register_fifo_parse_callback(BHY2_SENSOR_ID_GAMERV, parse_quaternion, (void*)&gamerv_accuracy, &bhy2);
		print_api_error(rslt, &bhy2);
		#endif

		rslt = bhy2_get_and_process_fifo(work_buffer, WORK_BUFFER_SIZE, &bhy2);
		print_api_error(rslt, &bhy2);
	}
	else
	{
		PDEBUG("Host interface not ready. Exiting\r\n");
		return 0;
	}

	/* Update the callback table to enable parsing of sensor data */
	rslt = bhy2_update_virtual_sensor_list(&bhy2);
	print_api_error(rslt, &bhy2);

	rslt = bhy2_set_virt_sensor_cfg(BHY2_SENSOR_ID_ACC_WU, sample_rate, report_latency_ms, &bhy2);
	print_api_error(rslt, &bhy2);
	PDEBUG("Enable %s at %.2fHz.\r\n", get_sensor_name(BHY2_SENSOR_ID_ACC_WU), sample_rate);

	rslt = bhy2_set_virt_sensor_cfg(BHY2_SENSOR_ID_GYRO_WU, sample_rate, report_latency_ms, &bhy2);
	print_api_error(rslt, &bhy2);
	PDEBUG("Enable %s at %.2fHz.\r\n", get_sensor_name(BHY2_SENSOR_ID_GYRO_WU), sample_rate);

	#if defined(BHI260AP_BMM150_9DOF)
	rslt = bhy2_set_virt_sensor_cfg(BHY2_SENSOR_ID_MAG_WU, sample_rate, report_latency_ms, &bhy2);
	print_api_error(rslt, &bhy2);
	PDEBUG("Enable %s at %.2fHz.\r\n", get_sensor_name(BHY2_SENSOR_ID_MAG_WU), sample_rate);
	
	rslt = bhy2_set_virt_sensor_cfg(BHY2_SENSOR_ID_ORI_WU, sample_rate, report_latency_ms, &bhy2);
	print_api_error(rslt, &bhy2);
	PDEBUG("Enable %s at %.2fHz.\r\n", get_sensor_name(BHY2_SENSOR_ID_ORI_WU), sample_rate);
	#elif defined(BHI260AP_6DOF)
	rslt = bhy2_set_virt_sensor_cfg(BHY2_SENSOR_ID_GAMERV, sample_rate, report_latency_ms, &bhy2);
	print_api_error(rslt, &bhy2);
	PDEBUG("Enable %s at %.2fHz.\r\n", get_sensor_name(BHY2_SENSOR_ID_GAMERV), sample_rate);
	#endif
	
	while (rslt == BHY2_OK)
	{
		#if defined(BHI260_FIFO_INTERRUPT_MODE)
		if(int1_flag == 1)
		{
			//PDEBUG("INT1 Triggered\r\n");
			int1_flag = 0;
		}
		if(bhi260_fifo_ready == BHY2_INTERRUPT_READY)
		{
			bhi260_fifo_ready = BHY2_INTERRUPT_IDLE;
			tick = HAL_GetTick();
			system_current_time = 1000*(int64_t)tick;
			//PDEBUG("before run bhy2_get_and_process_fifo(), %lld\r\n", system_current_time);
			/* Data from the FIFO is read and the relevant callbacks if registered are called */
			rslt = bhy2_get_and_process_fifo(work_buffer, WORK_BUFFER_SIZE, &bhy2);
			tick = HAL_GetTick();
			system_current_time = 1000*(int64_t)tick;
			//PDEBUG("after run bhy2_get_and_process_fifo(), %lld\r\n", system_current_time);
			print_api_error(rslt, &bhy2);
		}
		#elif defined(BHI260_FIFO_POLL_MODE)
		if (get_interrupt_status() == BHI260_GPIO_PIN_HIGH)	
		{
			/* Data from the FIFO is read and the relevant callbacks if registered are called */
			rslt = bhy2_get_and_process_fifo(work_buffer, WORK_BUFFER_SIZE, &bhy2);
			print_api_error(rslt, &bhy2);
		}
		#endif
	}

	return rslt;
}

void StartBHI260Task(void const * argument)
{
	PDEBUG("StartBHI260Task\r\n");
	
	#if defined(BHI260_FIFO_INTERRUPT_MODE)
	Enable_MCU_INT1_Pin();
	#elif defined(BHI260_FIFO_POLL_MODE)
	Enable_MCU_INT_Pin();
	#endif
	demo_sensor();
}

#endif
